// BonSrcDecoder.cpp: CBonSrcDecoder NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <TypeInfo.h>
#include "BonSrcDecoder.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


typedef IBonDriver* (PFCREATEBONDRIVER)(void);


//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

using namespace std;

CBonSrcDecoder::CBonSrcDecoder(IEventHandler *pEventHandler)
	: CMediaDecoder(pEventHandler, 0UL, 1UL)
	, m_pBonDriver(NULL)
	, m_pBonDriver2(NULL)
	, m_hStreamRecvThread(NULL)
	, m_TsStream(0x10000UL)
	, m_bKillSignal(true)
	, m_bIsPlaying(false)
	, m_dwLastError(BSDEC_NOERROR)
	, m_BitRate(0)
{

}

CBonSrcDecoder::~CBonSrcDecoder()
{
	CloseTuner();
}

void CBonSrcDecoder::Reset(void)
{
	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);

	EndStreamRecvThread();
	// ̃Xg[j
	if(m_bIsPlaying){
		// ̃Xg[j
		m_pBonDriver->PurgeTsStream();
	}

	// ʃfR[_Zbg
	CMediaDecoder::Reset();
	BeginStreamRecvThread();
}

const bool CBonSrcDecoder::InputMedia(CMediaData *pMediaData, const DWORD dwInputIndex)
{
	// \[XfR[_̂ߓ͂͏Ȃ
	return false;
}

const bool CBonSrcDecoder::OpenTuner(HMODULE hBonDrvDll)
{
	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);
	
	// I[v`FbN
	if(m_pBonDriver){
		m_dwLastError = BSDEC_ALREADYOPEN;
		return false;
		}

	// hCo|C^̎擾
	PFCREATEBONDRIVER *pf=(PFCREATEBONDRIVER*)GetProcAddress(hBonDrvDll,"CreateBonDriver");
	if(!pf){
		m_dwLastError = BSDEC_DRIVERERROR;
		return false;
		}

	// hCoCX^X
	if(!(m_pBonDriver = pf())){
		m_dwLastError = BSDEC_TUNERERROR;
		return false;
		}

	// IBonDriver2C^tF[X擾
	try{
		m_pBonDriver2 = dynamic_cast<IBonDriver2 *>(m_pBonDriver);
		}
	catch(std::__non_rtti_object){
		m_pBonDriver2 = NULL;
		}

	// `[iJ
	if(!m_pBonDriver->OpenTuner()){
		CloseTuner();
		m_dwLastError = BSDEC_TUNERERROR;
		return false;
		}
	
	// `lZbg
	if(m_pBonDriver2){
		// IBonDriver2
		if(!m_pBonDriver2->SetChannel(0UL, 0UL)){
			CloseTuner();
			m_dwLastError = BSDEC_TUNERERROR;
			return false;
			}
		}
	else{
		// IBonDriver
		if(!m_pBonDriver->SetChannel(13U)){
			CloseTuner();
			m_dwLastError = BSDEC_TUNERERROR;
			return false;
			}
		}

	// Xg[MXbhN
	DWORD dwThreadID = 0UL;
	m_bKillSignal = false;
	m_bIsPlaying = false;

	if(!(m_hStreamRecvThread = ::CreateThread(NULL, 0UL, CBonSrcDecoder::StreamRecvThread, (LPVOID)this, 0UL, &dwThreadID))){
		CloseTuner();
		m_dwLastError = BSDEC_INTERNALERROR;
		return false;
		}

	m_dwLastError = BSDEC_NOERROR;

	ResetBitRate();
	return true;
}

const bool CBonSrcDecoder::CloseTuner(void)
{
	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);

	// Xg[~
	m_bIsPlaying = false;
	m_bKillSignal = true;

	if(m_hStreamRecvThread){
		// Xg[MXbh~
		if(::WaitForSingleObject(m_hStreamRecvThread, 1000UL) != WAIT_OBJECT_0){
			// XbhI
			::TerminateThread(m_hStreamRecvThread, 0UL);
			}

		::CloseHandle(m_hStreamRecvThread);
		m_hStreamRecvThread = NULL;
		}

	if(m_pBonDriver){
		// `[i
		m_pBonDriver->CloseTuner();

		// hCoCX^XJ
		m_pBonDriver->Release();
		m_pBonDriver = NULL;		
		m_pBonDriver2 = NULL;		
		}

	m_dwLastError = BSDEC_NOERROR;

	return true;
}

const bool CBonSrcDecoder::Play(void)
{
	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);
	
	if(!m_pBonDriver){
		// `[iJĂȂ
		m_dwLastError = BSDEC_TUNERNOTOPEN;
		return false;
		}

	if(m_bIsPlaying){
		// ɍĐ
		m_dwLastError = BSDEC_ALREADYPLAYING;
		return false;
		}

	EndStreamRecvThread();
	// ̃Xg[j
	m_pBonDriver->PurgeTsStream();

	// ʃfR[_Zbg
	CMediaDecoder::Reset();
	BeginStreamRecvThread();

	// Xg[ĐԂɂ
	m_bIsPlaying = true;
	m_dwLastError = BSDEC_NOERROR;

	return true;
}

const bool CBonSrcDecoder::Stop(void)
{
	if(!m_bIsPlaying){
		// Xg[͍ĐłȂ
		m_dwLastError = BSDEC_NOTPLAYING;
		return false;
		}

	m_bIsPlaying = false;
	m_dwLastError = BSDEC_NOERROR;

	return true;
}

const bool CBonSrcDecoder::SetChannel(const BYTE byChannel)
{
	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);

	if(!m_pBonDriver){
		// `[iJĂȂ
		m_dwLastError = BSDEC_TUNERNOTOPEN;
		return false;
		}

	EndStreamRecvThread();
	// ̃Xg[j
	m_pBonDriver->PurgeTsStream();

	// `lύX
	if(!m_pBonDriver->SetChannel(byChannel)){
		m_dwLastError = BSDEC_TUNERERROR;
		return false;
		}

	// ʃfR[_Zbg
	CMediaDecoder::Reset();
	BeginStreamRecvThread();

	m_dwLastError = BSDEC_NOERROR;
	
	return true;
}

const bool CBonSrcDecoder::SetChannel(const DWORD dwSpace, const DWORD dwChannel)
{
	if(!IsBonDriver2())return SetChannel((BYTE)dwChannel + 13U);

	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);

	if(!m_pBonDriver2){
		// `[iJĂȂ
		m_dwLastError = BSDEC_TUNERNOTOPEN;
		return false;
		}

	EndStreamRecvThread();
	// ̃Xg[j
	m_pBonDriver2->PurgeTsStream();

	// `lύX
	if(!m_pBonDriver2->SetChannel(dwSpace, dwChannel)){
		m_dwLastError = BSDEC_TUNERERROR;
		return false;
		}

	// ʃfR[_Zbg
	CMediaDecoder::Reset();
	BeginStreamRecvThread();

	m_dwLastError = BSDEC_NOERROR;
	
	return true;
}

const float CBonSrcDecoder::GetSignalLevel(void)
{
	if(!m_pBonDriver){
		// `[iJĂȂ
		m_dwLastError = BSDEC_TUNERNOTOPEN;
		return 0.0f;
		}

	m_dwLastError = BSDEC_NOERROR;

	// MxԂ
	return m_pBonDriver->GetSignalLevel();
}

const bool CBonSrcDecoder::IsBonDriver2(void) const
{
	// IBonDriver2C^tF[X̎gpۂԂ
	return (m_pBonDriver2)? true : false;
}

LPCTSTR CBonSrcDecoder::GetSpaceName(const DWORD dwSpace) const
{
	// `[jOԖԂ
	return (IsBonDriver2())? m_pBonDriver2->EnumTuningSpace(dwSpace) : NULL;
}

LPCTSTR CBonSrcDecoder::GetChannelName(const DWORD dwSpace, const DWORD dwChannel) const
{
	// `lԂ
	return (IsBonDriver2())? m_pBonDriver2->EnumChannelName(dwSpace, dwChannel) : NULL;
}

const DWORD CBonSrcDecoder::GetLastError(void) const
{
	// ŌɔG[Ԃ
	return m_dwLastError;
}

const bool CBonSrcDecoder::PurgeStream(void)
{
	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);

	if(!m_pBonDriver){
		// `[iJĂȂ
		m_dwLastError = BSDEC_TUNERNOTOPEN;
		return false;
		}

	EndStreamRecvThread();
	// ̃Xg[j
	m_pBonDriver->PurgeTsStream();

	// ʃfR[_Zbg
	CMediaDecoder::Reset();
	BeginStreamRecvThread();
	
	m_dwLastError = BSDEC_NOERROR;

	return true;
}

void CBonSrcDecoder::OnTsStream(BYTE *pStreamData, DWORD dwStreamSize)
{
	CBlockLock Lock(&m_CriticalLock);
	TRACE(TEXT("Lock %d\n"),__LINE__);

	if(!m_bIsPlaying)return;

	// ŏʃfR[_ɓ͂
	m_TsStream.SetData(pStreamData, dwStreamSize);
	OutputMedia(&m_TsStream);
}

DWORD WINAPI CBonSrcDecoder::StreamRecvThread(LPVOID pParam)
{
	CBonSrcDecoder *pThis = static_cast<CBonSrcDecoder *>(pParam);

	BYTE *pStreamData = NULL;
	DWORD dwStreamSize = 0UL;
	DWORD dwStreamRemain = 0UL;
	DWORD TotalSize=0;

	// `[iTSf[^oXbh
	while(!pThis->m_bKillSignal){
		// ȗ̂߃|[O̗p
		do{
			if(pThis->m_pBonDriver->GetTsStream(&pStreamData, &dwStreamSize, &dwStreamRemain)){
				if(pStreamData && dwStreamSize){
					pThis->OnTsStream(pStreamData, dwStreamSize);
					}
				}
			TotalSize+=dwStreamSize;
			}
		while(dwStreamRemain);

		// rbg[gvZ
		DWORD Now=GetTickCount();
		if (Now>pThis->m_BitRateTime) {
			if (Now-pThis->m_BitRateTime>=1000) {
				pThis->m_BitRate=(TotalSize)*1000/(Now-pThis->m_BitRateTime);
				pThis->m_BitRateTime=Now;
				TotalSize=0;
			}
		} else {
			pThis->m_BitRateTime=Now;
			TotalSize=0;
		}

		// EFCg(24MbpsƂĎ̃f[^܂Ŗ15ms)
		::Sleep(1UL);
		}

	return 0UL;
}




bool CBonSrcDecoder::BeginStreamRecvThread()
{
	DWORD dwThreadID;

	if (m_hStreamRecvThread!=NULL)
		return false;
	m_hStreamRecvThread = ::CreateThread(NULL, 0UL, CBonSrcDecoder::StreamRecvThread, (LPVOID)this, 0UL, &dwThreadID);
	if (m_hStreamRecvThread==NULL)
		return false;
	ResetBitRate();
	return true;
}


bool CBonSrcDecoder::EndStreamRecvThread()
{
	if (m_hStreamRecvThread) {
		m_bKillSignal=true;
		if (::WaitForSingleObject(m_hStreamRecvThread, 3000UL) != WAIT_OBJECT_0) {
			::TerminateThread(m_hStreamRecvThread, 1UL);
		}
		::CloseHandle(m_hStreamRecvThread);
		m_hStreamRecvThread = NULL;
		m_bKillSignal=false;
	}
	return true;
}


LPCTSTR CBonSrcDecoder::GetTunerName() const
{
	if (m_pBonDriver2==NULL)
		return NULL;
	return m_pBonDriver2->GetTunerName();
}


DWORD CBonSrcDecoder::GetBitRate() const
{
	return m_BitRate;
}

void CBonSrcDecoder::ResetBitRate()
{
	m_BitRateTime=GetTickCount();
	m_BitRate=0;

}
